{
ZZZZZZ 5.0, a chess program.
Copyright (C) 1998 G. Wiesenekker
E-mail: gijsbert.wiesenekker@wxs.nl
http://home.wxs.nl/~gijsbert.wiesenekker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}
unit Lists;

{$MODE Delphi}

interface

procedure SetLists;
procedure CompactLists;

implementation

uses SysUtils, Chess, Rand, Keys, Score;

const BoardMap: array[0..7, 0..7] of BoardRange =
	((A8, B8, C8, D8, E8, F8, G8, H8),
	 (A7, B7, C7, D7, E7, F7, G7, H7),
	 (A6, B6, C6, D6, E6, F6, G6, H6),
	 (A5, B5, C5, D5, E5, F5, G5, H5),
	 (A4, B4, C4, D4, E4, F4, G4, H4),
	 (A3, B3, C3, D3, E3, F3, G3, H3),
	 (A2, B2, C2, D2, E2, F2, G2, H2),
	 (A1, B1, C1, D1, E1, F1, G1, H1));
var
	HashKeys: array[BoardColourEnum, BoardKindEnum] of KeyArray;

procedure InitHashKeys;
var
	Colour: BoardColourEnum;
  Kind: BoardKindEnum;
  i: BoardRange;
  Sum, CheckSum: Key;
  KeyPointer: ^KeyArray;
begin
  for Colour := Low(BoardColourEnum) to High(BoardColourEnum) do
  begin
  	for Kind := Low(BoardKindEnum) to High(BoardKindEnum) do
    begin
    	Sum := 0;
    	for i:= Low(BoardRange) to High(BoardRange) do
      begin
				HashKeys[Colour, Kind, i] := GenKey;
        Sum := Sum xor HashKeys[Colour, Kind, i];
      end;
      KeyPointer := @HashKeys[Colour, Kind];
      CheckSum := 0;
    	for i:= Low(BoardRange) to High(BoardRange) do
        CheckSum := CheckSum xor KeyPointer[i];
      if (CheckSum <> Sum) then
      	raise EChessError.Create('CheckSum error');
    end;
  end;
  WhiteShortCastleKey := GenKey;
  WhiteLongCastleKey := GenKey;
  BlackShortCastleKey := GenKey;
  BlackLongCastleKey := GenKey;
  for i:= Low(BoardRange) to High(BoardRange) do
    EnPassantKey[i] := GenKey;
end;

procedure InitLists;
var
	Colour: BoardColourEnum;
  Kind: BoardKindEnum;
  List: ListRange;
begin
  for Colour := Low(BoardColourEnum) to High(BoardColourEnum) do
  begin
  	for Kind := Low(BoardKindEnum) to High(BoardKindEnum) do
    begin
    	PieceCount[Colour, Kind] := 0;
    	for List := Low(ListRange) to High(ListRange) do
      begin
    		with PieceList[Colour, Kind, List] do
        begin
        	ListColour := Colour;
          ListKind := Kind;
          ListWhere := Low(ExtBoardRange);
          ListMoved := False;
          ListKey := @HashKeys[Colour, Kind];
        end;
    	end;
    end;
  end;
end;

procedure CompactLists;
var
	iColour: BoardColourEnum;
  iKind: BoardKindEnum;
  iList, jList: ExtListRange;
begin
  for iColour := Low(BoardColourEnum) to High(BoardColourEnum) do
  begin
  	for iKind := Low(BoardKindEnum) to High(BoardKindEnum) do
    begin
    	iList := 1;
      while iList <= PieceCount[iColour, iKind] do
      begin
        if PieceList[iColour, iKind, iList].ListWhere = Low(ExtBoardRange) then
        begin
          for jList := iList to PieceCount[iColour, iKind] - 1 do
          begin
          	PieceList[iColour, iKind, jList] := PieceList[iColour, iKind, jList + 1];
            if PieceList[iColour, iKind, jList].ListWhere <> Low(ExtBoardRange) then
            begin
          		Board[PieceList[iColour, iKind, jList].ListWhere] :=
          			@PieceList[iColour, iKind, jList];
            end;
        	end;
          Dec(PieceCount[iColour, iKind]);
        end
        else
        begin
        	Inc(iList);
        end
      end;
    end;
  end;
  CheckHashKey;
  CheckMaterialScore;
end;

procedure ClearBoard;
var
	i: BoardRange;
	Col, Row: Integer;
begin
  for i := Low(BoardRange) to High(BoardRange) do
  	Board[i] := @SideSquare;
  for Col := 0 to 7 do
  	for Row := 0 to 7 do
    	Board[BoardMap[Row, Col]] := @EmptySquare;
end;

procedure ClearLists;
var
	Colour: BoardColourEnum;
  Kind: BoardKindEnum;
begin
  for Colour := Low(BoardColourEnum) to High(BoardColourEnum) do
  begin
  	for Kind := Low(BoardKindEnum) to High(BoardKindEnum) do
    begin
    	PieceCount[Colour, Kind] := 0;
    end;
  end;
end;

procedure SetLists;
var
	Col, Row: Integer;
  Colour: BoardColourEnum;
  Kind: BoardKindEnum;
  Found: Boolean;
begin
	ClearLists;
  ClearBoard;
  for Col := 0 to 7 do
  begin
  	for Row := 0 to 7 do
    begin
    	Found := False;
      if (GameBoardGrid[Row, Col] = wP) then
      begin
      	Colour := White;
        Kind := Pawn;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = wN) then
      begin
      	Colour := White;
        Kind := Knight;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = wB) then
      begin
      	Colour := White;
        Kind := Bishop;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = wR) then
      begin
      	Colour := White;
        Kind := Rook;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = wQ) then
      begin
      	Colour := White;
        Kind := Queen;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = wK) then
      begin
      	Colour := White;
        Kind := King;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = bP) then
      begin
      	Colour := Black;
        Kind := Pawn;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = bN) then
      begin
      	Colour := Black;
        Kind := Knight;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = bB) then
      begin
      	Colour := Black;
        Kind := Bishop;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = bR) then
      begin
      	Colour := Black;
        Kind := Rook;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = bQ) then
      begin
      	Colour := Black;
        Kind := Queen;
        Found := True;
      end
      else if (GameBoardGrid[Row, Col] = bK) then
      begin
      	Colour := Black;
        Kind := King;
        Found := True;
      end;
      if (Found) then
      begin
      	Inc(PieceCount[Colour, Kind]);
      	with PieceList[Colour, Kind, PieceCount[Colour, Kind]] do
      	begin
        	ListWhere := BoardMap[Row, Col];
          if GameBoardGrid[Row, Col] = StartBoardGrid[Row, Col] then
          	ListMoved := False
          else
          	ListMoved := True;
          ListCaptured := False;
          Board[ListWhere] :=
          	@PieceList[Colour, Kind, PieceCount[Colour, Kind]];
      	end;
      end;
    end;
  end;
  HashKey := HashKeyFromLists;
  SetMaterialScore;
  CheckHashKey;
  CheckMaterialScore;
  Node := 1;
  NodeList[Node].NodeHashKey := HashKey;
  NodeList[Node].NodeMaterialScore := MaterialScore;
  NodeList[Node].NodeEnPassant := Low(ExtBoardRange);
  SetFullHashKey;
end;

initialization

InitHashKeys;
InitLists;

end.
