{
ZZZZZZ 5.0, a chess program.
Copyright (C) 1998 G. Wiesenekker
E-mail: gijsbert.wiesenekker@wxs.nl
http://home.wxs.nl/~gijsbert.wiesenekker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}
unit Moves;

{$MODE Delphi}

interface

uses Chess, Utils;

function ReturnCheck(MyColour, YourColour: BoardColourEnum): CheckEnum;
procedure GenerateMoves(MyColour, YourColour: BoardColourEnum; Option: MovesEnum;
	var nMoves: ExtMoveRange; var Moves: MoveList);
function MoveToLongString(TheMove: MoveRecord): AnsiString;
function MoveToDragDropString(TheMove: MoveRecord): AnsiString;
procedure DoMove(MyColour, YourColour: BoardColourEnum;
	TheMove: MoveRecord);
procedure UndoMove(MyColour, YourColour: BoardColourEnum;
	TheMove: MoveRecord);
procedure GenerateLegalMoves(MyColour, YourColour: BoardColourEnum;
	var nMoves: ExtMoveRange; var Moves: MoveList);
function MoveOk(nMoves: ExtMoveRange; var Moves: MoveList;
	TheMove: MoveRecord): Boolean;
const NullMove: MoveRecord = (MoveFrom: 1; MoveTo: 1; MoveKind: [Null]);
const InvalidMove: MoveRecord = (MoveFrom: 1; MoveTo: 1; MoveKind: [Invalid]);

implementation

uses Keys, Score, Main;

function ReturnAttacked(MyColour, YourColour: BoardColourEnum;
	iBoard: BoardRange): Boolean;
var
	jBoard: ExtBoardRange;
  iDir: Integer;
begin
	Result := True;
  if PieceCount[YourColour, Pawn] > 0 then
  begin
  	jBoard := iBoard + PawnCaptureDir1[MyColour];
    if (Board[jBoard].ListKind = Pawn) and (Board[jBoard].ListColour = YourColour) then
    	Exit;
  	jBoard := iBoard + PawnCaptureDir2[MyColour];
    if (Board[jBoard].ListKind = Pawn) and (Board[jBoard].ListColour = YourColour) then
    	Exit;
  end;
  if PieceCount[YourColour, Knight] > 0 then
  begin
  	for iDir := 1 to 8 do
    begin
    	jBoard := iBoard + KnightDir[iDir];
    	if (Board[jBoard].ListKind = Knight) and (Board[jBoard].ListColour = YourColour) then
    		Exit;
    end;
  end;
  if (PieceCount[YourColour, Bishop] > 0) or (PieceCount[YourColour, Queen] > 0) then
  begin
  	for iDir := 1 to 4 do
    begin
    	jBoard := iBoard + BishopDir[iDir];
      while Board[jBoard].ListKind = Empty do
      	Inc(jBoard, BishopDir[iDir]);
    	if ((Board[jBoard].ListKind = Bishop) or (Board[jBoard].ListKind = Queen)) and
      	  (Board[jBoard].ListColour = YourColour) then
    		Exit;
    end;
  end;
  if (PieceCount[YourColour, Rook] > 0) or (PieceCount[YourColour, Queen] > 0) then
  begin
  	for iDir := 1 to 4 do
    begin
    	jBoard := iBoard + RookDir[iDir];
      while Board[jBoard].ListKind = Empty do
      	Inc(jBoard, RookDir[iDir]);
    	if ((Board[jBoard].ListKind = Rook) or (Board[jBoard].ListKind = Queen)) and
      	  (Board[jBoard].ListColour = YourColour) then
    		Exit;
    end;
  end;
  for iDir := 1 to 8 do
  begin
  	jBoard := iBoard + KingDir[iDir];
    if (Board[jBoard].ListKind = King) and (Board[jBoard].ListColour = YourColour) then
    	Exit;
  end;
  Result := False;
end;

function ReturnCheck(MyColour, YourColour: BoardColourEnum): CheckEnum;
begin
	if ReturnAttacked(MyColour, YourColour, PieceList[MyColour, King, 1].ListWhere) then
  	Result := Check
  else
  	Result := NoCheck;
end;

procedure GenerateMoves(MyColour, YourColour: BoardColourEnum; Option: MovesEnum;
	var nMoves: ExtMoveRange; var Moves: MoveList);
var
	iList: ExtListRange;
  iBoard: ExtBoardRange;
  Temp: MoveList;
  nTemp, iTemp: ExtMoveRange;

  procedure AddMove(jBoard: BoardRange; Kind: MoveKindSet);
  begin
  	if not (Normal in Kind) then
    begin
    	Inc(nMoves);
    	with Moves[nMoves] do
    	begin
      	MoveFrom := iBoard;
      	MoveTo := jBoard;
      	MoveKind := Kind;
    	end;
    end
    else
    begin
    	Inc(nTemp);
    	with Temp[nTemp] do
    	begin
      	MoveFrom := iBoard;
      	MoveTo := jBoard;
      	MoveKind := Kind;
    	end;
    end;
  end;

  procedure AddPawnMove(jBoard: BoardRange; Kind: MoveKindSet);
  begin
  	if RankNumber[MyColour, jBoard] < 8 then
    	AddMove(jBoard, Kind)
    else
    begin
    	AddMove(jBoard, Kind + [Promotion, PromotionQueen]);
    	AddMove(jBoard, Kind + [Promotion, PromotionRook]);
    	AddMove(jBoard, Kind + [Promotion, PromotionBishop]);
    	AddMove(jBoard, Kind + [Promotion, PromotionKnight]);
  	end;
  end;

  procedure GeneratePawnMoves(Dir: Integer; Dir1, Dir2: Integer);
  var
  	jBoard: BoardRange;
  begin
  	jBoard := iBoard + Dir1;
    if Board[jBoard].ListColour = YourColour then
			AddPawnMove(jBoard, [Capture]);
  	jBoard := iBoard + Dir2;
    if Board[jBoard].ListColour = YourColour then
			AddPawnMove(jBoard, [Capture]);
  	jBoard := iBoard + Dir;
    if Board[jBoard].ListKind = Empty then
    begin
			AddPawnMove( jBoard, [Normal]);
    	if RankNumber[MyColour, iBoard] = 2 then
    	begin
    		jBoard := jBoard + Dir;
        if Board[jBoard].ListKind = Empty then
        	AddMove(jBoard, [Normal, DoublePawn]);
    	end;
    end;
  end;

	procedure GenerateKnightKingMoves(Dir: DirArray; NDir: DirRange);
  var
  	jBoard: ExtBoardRange;
  	iDir: Integer;
  begin
    for iDir := 1 to NDir do
    begin
    	jBoard := iBoard + Dir[iDir];
      if Board[jBoard].ListKind = Empty then
				AddMove(jBoard, [Normal])
      else if Board[jBoard].ListColour = YourColour then
				AddMove(jBoard, [Capture]);
    end;
  end;

	procedure GenerateBishopRookQueenMoves(Dir: DirArray; NDir: DirRange);
  var
  	jBoard: BoardRange;
  	iDir: Integer;
  begin
    for iDir := 1 to NDir do
    begin
    	jBoard := iBoard + Dir[iDir];
    	while Board[jBoard].ListKind = Empty do
      begin
				AddMove(jBoard, [Normal]);
        Inc(jBoard, Dir[iDir]);
      end;
      if Board[jBoard].ListColour = YourColour then
				AddMove(jBoard, [Capture]);
    end;
  end;

  procedure AddCastle(KingFrom, KingTo, RookFrom, RookTo: BoardRange;
  	RookEmpty: ExtBoardRange = Low(ExtBoardRange));
  begin
		if Board[KingFrom].ListMoved then
    	Exit;
    if Board[RookFrom].ListMoved then
    	Exit;
    if Board[KingTo].ListKind <> Empty then
    	Exit;
    if Board[RookTo].ListKind <> Empty then
    	Exit;
    //only for long castle
    if RookEmpty <> Low(ExtBoardRange) then
    begin
    	if Board[RookEmpty].ListKind <> Empty then
      	Exit;
    end;
    //final check
    if ReturnAttacked(MyColour, YourColour, KingFrom) then
    	Exit;
    if ReturnAttacked(MyColour, YourColour, KingTo) then
    	Exit;
    if ReturnAttacked(MyColour, YourColour, RookTo) then
    	Exit;
    iBoard := KingFrom;
    if RookEmpty = Low(ExtBoardRange) then
    	AddMove(KingTo, [ShortCastle])
    else
    	AddMove(KingTo, [LongCastle]);
  end;

begin
	nTemp := 0;
	nMoves := 0;
  if NodeList[Node].NodeEnPassant <> Low(ExtBoardRange) then
  begin
    iBoard := NodeList[Node].NodeEnPassant - 1;
    if (Board[iBoard].ListColour = MyColour) and (Board[iBoard].ListKind = Pawn) then
    	AddMove(NodeList[Node].NodeEnPassant + PawnMoveDir[MyColour], [Capture, EnPassant]);
    iBoard := NodeList[Node].NodeEnPassant + 1;
    if (Board[iBoard].ListColour = MyColour) and (Board[iBoard].ListKind = Pawn) then
    	AddMove(NodeList[Node].NodeEnPassant + PawnMoveDir[MyColour], [Capture, EnPassant]);
  end;
	for iList := 1 to PieceCount[MyColour, Pawn] do
  begin
  	iBoard := PieceList[MyColour, Pawn, iList].ListWhere;
    if iBoard = Low(ExtBoardRange) then continue;
    GeneratePawnMoves(PawnMoveDir[MyColour],
    	PawnCaptureDir1[MyColour], PawnCaptureDir2[MyColour]);
  end;
	for iList := 1 to PieceCount[MyColour, Knight] do
  begin
  	iBoard := PieceList[MyColour, Knight, iList].ListWhere;
    if iBoard = Low(ExtBoardRange) then continue;
    GenerateKnightKingMoves(KnightDir, 8);
  end;
	for iList := 1 to PieceCount[MyColour, Bishop] do
  begin
  	iBoard := PieceList[MyColour, Bishop, iList].ListWhere;
    if iBoard = Low(ExtBoardRange) then continue;
    GenerateBishopRookQueenMoves(BishopDir, 4);
  end;
	for iList := 1 to PieceCount[MyColour, Rook] do
  begin
  	iBoard := PieceList[MyColour, Rook, iList].ListWhere;
    if iBoard = Low(ExtBoardRange) then continue;
    GenerateBishopRookQueenMoves(RookDir, 4);
  end;
	for iList := 1 to PieceCount[MyColour, Queen] do
  begin
  	iBoard := PieceList[MyColour, Queen, iList].ListWhere;
    if iBoard = Low(ExtBoardRange) then continue;
    GenerateBishopRookQueenMoves(QueenDir, 8);
  end;
	for iList := 1 to PieceCount[MyColour, King] do
  begin
  	iBoard := PieceList[MyColour, King, iList].ListWhere;
    if iBoard = Low(ExtBoardRange) then continue;
    GenerateKnightKingMoves(KingDir, 8);
	end;
  if Option = AllMoves then
  begin
  	if MyColour = White then
    begin
    	AddCastle(E1, G1, H1, F1);
    	AddCastle(E1, C1, A1, D1, B1);
    end
    else
    begin
    	AddCastle(E8, G8, H8, F8);
    	AddCastle(E8, C8, A8, D8, B8);
    end;
  	for iTemp := 1 to nTemp do
    begin
    	Inc(nMoves);
      Moves[nMoves] := Temp[iTemp];
    end;
  end;
  Moves[MovesMax] := NulLMove;
end;

function MoveToLongString(TheMove: MoveRecord): AnsiString;
begin
	with TheMove do
  begin
  	if Normal in MoveKind then
			Result := BoardIndexToString(TheMove.MoveFrom) + '-' +
  			BoardIndexToString(TheMove.MoveTo)
  	else if Capture in MoveKind then
			Result := BoardIndexToString(TheMove.MoveFrom) + 'x' +
  			BoardIndexToString(TheMove.MoveTo)
    else if ShortCastle in MoveKind then
    	Result := 'o-o'
    else if LongCastle in MoveKind then
    	Result := 'o-o-o'
    else if MoveKind = [Null] then
    	Result := 'Null'
    else if MoveKind = [Invalid] then
    	Result := 'Invalid';
    if Promotion in MoveKind then
    begin
      if PromotionQueen in MoveKind then
        Result := Result + 'Q'
      else if PromotionRook in MoveKind then
        Result := Result + 'R'
      else if PromotionBishop in MoveKind then
        Result := Result + 'B'
      else if PromotionKnight in MoveKind then
        Result := Result + 'N'
      else
      	raise EChessError.Create('Invalid Promotion Kind');
    end;
  end;
end;

function MoveToDragDropString(TheMove: MoveRecord): AnsiString;
begin
	with TheMove do
		Result := BoardIndexToString(TheMove.MoveFrom) +
			BoardIndexToString(TheMove.MoveTo)
end;

procedure DoMove(MyColour, YourColour: BoardColourEnum;
	TheMove: MoveRecord);
var
	iBoard: ExtBoardRange;
  MoveCapture: ExtBoardRange;

  procedure DoCastleMove(MoveFrom, MoveTo: BoardRange);
  begin
  	HashKey := HashKey xor Board[MoveFrom].ListKey[MoveFrom];
    Board[MoveTo] := Board[MoveFrom];
    Board[MoveTo].ListWhere := MoveTo;
    Board[MoveTo].ListMoved := True;
    Board[MoveFrom] := @EmptySquare;
  	HashKey := HashKey xor Board[MoveTo].ListKey[MoveTo];
  end;
begin
	with NodeList[Node] do
  begin
		NodeHashKey := HashKey;
    NodeMaterialScore := MaterialScore;
		NodeCapture := False;
  end;
	with TheMove do
  begin
  	if MoveKind = [Null] then
    	NodeList[Node].NodeNull := True
    else
    	NodeList[Node].NodeNull := False;
  	if Movekind <> [Null] then
    begin
    	if MoveKind = [ShortCastle] then
      begin
      	if MyColour = White then
        begin
          DoCastleMove(E1, G1);
          DoCastleMove(H1, F1);
        end
        else
        begin
          DoCastleMove(E8, G8);
          DoCastleMove(H8, F8);
        end;
      end
      else if MoveKind = [LongCastle] then
      begin
      	if MyColour = White then
        begin
          DoCastleMove(E1, C1);
          DoCastleMove(A1, D1);
        end
        else
        begin
          DoCastleMove(E8, C8);
          DoCastleMove(A8, D8);
        end;
      end
      else
      begin
        NodeList[Node].NodeListMoved := Board[MoveFrom].ListMoved;
        NodeList[Node].NodeListCaptured := Board[MoveFrom].ListCaptured;
        if Capture in MoveKind then
        begin
        	if EnPassant in MoveKind then
          	MoveCapture := MoveTo - PawnMoveDir[MyColour]
          else
          	MoveCapture := MoveTo;
          with NodeList[Node] do
          begin
            NodeCapture := True;
            TheCapture := Board[MoveCapture];
            Board[MoveFrom].ListCaptured := True;
          end;
          with Board[MoveCapture]^ do
          begin
            HashKey := HashKey xor ListKey[MoveCapture];
            Dec(MaterialScore[YourColour], MaterialValue[ListKind]);
            ListWhere := Low(ExtBoardRange);
          end;
          Board[MoveCapture] := @EmptySquare;
        end;
        with Board[MoveFrom]^ do
        begin
          HashKey := HashKey xor ListKey[MoveFrom];
        end;
        if Promotion in MoveKind then
        begin
          NodeList[Node].ThePromotion := Board[MoveFrom];
          Board[MoveFrom].ListWhere := Low(ExtBoardRange);
          Dec(MaterialScore[MyColour], MaterialValue[Pawn]);
          if PromotionQueen in MoveKind then
          begin
            Inc(PieceCount[MyColour, Queen]);
            Board[MoveTo] := @PieceList[MyColour, Queen, PieceCount[MyColour, Queen]];
            Inc(MaterialScore[MyColour], MaterialValue[Queen]);
          end
          else if PromotionRook in MoveKind then
          begin
            Inc(PieceCount[MyColour, Rook]);
            Board[MoveTo] := @PieceList[MyColour, Rook, PieceCount[MyColour, Rook]];
            Inc(MaterialScore[MyColour], MaterialValue[Rook]);
          end
          else if PromotionBishop in MoveKind then
          begin
            Inc(PieceCount[MyColour, Bishop]);
            Board[MoveTo] := @PieceList[MyColour, Bishop, PieceCount[MyColour, Bishop]];
            Inc(MaterialScore[MyColour], MaterialValue[Bishop]);
          end
          else if PromotionKnight in MoveKind then
          begin
            Inc(PieceCount[MyColour, Knight]);
            Board[MoveTo] := @PieceList[MyColour, Knight, PieceCount[MyColour, Knight]];
            Inc(MaterialScore[MyColour], MaterialValue[Knight]);
          end
        end
        else
        begin
          Board[MoveTo] := Board[MoveFrom];
        end;
        Board[MoveFrom] := @EmptySquare;
        with Board[MoveTo]^ do
        begin
          HashKey := HashKey xor ListKey^[MoveTo];
          ListWhere := MoveTo;
          ListMoved := True;
        end;
      end;
      if (Debug) then
      begin
        CheckHashKey;
        CheckMaterialScore;
      end;
    end;
  end;
  Inc(Node);
  NodeList[Node].NodeEnPassant := Low(ExtBoardRange);
  if DoublePawn in TheMove.MoveKind then
  begin
    iBoard := TheMove.MoveTo - 1;
    if (Board[iBoard].ListColour = YourColour) and (Board[iBoard].ListKind = Pawn) then
      NodeList[Node].NodeEnPassant := TheMove.MoveTo;
    iBoard := TheMove.MoveTo + 1;
    if (Board[iBoard].ListColour = YourColour) and (Board[iBoard].ListKind = Pawn) then
      NodeList[Node].NodeEnPassant := TheMove.MoveTo;
  end;
  SetFullHashKey;
end;

procedure UndoMove(MyColour, YourColour: BoardColourEnum;
	TheMove: MoveRecord);
var
	MoveCapture: ExtBoardRange;
  
  procedure UndoCastleMove(MoveFrom, MoveTo: BoardRange);
  begin
  	HashKey := HashKey xor Board[MoveFrom].ListKey[MoveFrom];
    Board[MoveTo] := Board[MoveFrom];
    Board[MoveTo].ListWhere := MoveTo;
    Board[MoveTo].ListMoved := False;
    Board[MoveFrom] := @EmptySquare;
  	HashKey := HashKey xor Board[MoveTo].ListKey[MoveTo];
  end;
begin
	Dec(Node);
	with TheMove do
  begin
  	if MoveKind <> [Null] then
    begin
    	if MoveKind = [ShortCastle] then
      begin
      	if MyColour = White then
        begin
          UndoCastleMove(G1, E1);
          UndoCastleMove(F1, H1);
        end
        else
        begin
          UndoCastleMove(G8, E8);
          UndoCastleMove(F8, H8);
        end;
      end
      else if MoveKind = [LongCastle] then
      begin
      	if MyColour = White then
        begin
          UndoCastleMove(C1, E1);
          UndoCastleMove(D1, A1);
        end
        else
        begin
          UndoCastleMove(C8, E8);
          UndoCastleMove(D8, A8);
        end;
      end
      else
      begin
        with Board[MoveTo]^ do
        begin
          HashKey := HashKey xor ListKey^[MoveTo];
        end;
        if Promotion in MoveKind then
        begin
          Board[MoveFrom] := NodeList[Node].ThePromotion;
          if PromotionQueen in MoveKind then
          begin
            Dec(PieceCount[MyColour, Queen]);
          end
          else if PromotionRook in MoveKind then
          begin
            Dec(PieceCount[MyColour, Rook]);
          end
          else if PromotionBishop in MoveKind then
          begin
            Dec(PieceCount[MyColour, Bishop]);
          end
          else if PromotionKnight in MoveKind then
          begin
            Dec(PieceCount[MyColour, Knight]);
          end
        end
        else
          Board[MoveFrom] := Board[MoveTo];
        with Board[MoveFrom]^ do
        begin
          HashKey := HashKey xor ListKey[MoveFrom];
          ListWhere := MoveFrom;
          ListMoved := NodeList[Node].NodeListMoved;
          ListCaptured := NodeList[Node].NodeListCaptured;
        end;
        Board[MoveTo] := @EmptySquare;
        if Capture in MoveKind then
        begin
        	if EnPassant in MoveKind then
          	MoveCapture := MoveTo - PawnMoveDir[MyColour]
          else
          	MoveCapture := MoveTo;
          Board[MoveCapture] := NodeList[Node].TheCapture;
          with Board[MoveCapture]^ do
          begin
            HashKey := HashKey xor ListKey[MoveCapture];
            ListWhere := MoveCapture;
          end;
        end;
      end;
    end;
  end;
  with NodeList[Node] do
  begin
  	HashKey := NodeHashKey;
  	MaterialScore := NodeMaterialScore;
  end;
	if (Debug) then
  begin
    CheckHashKey;
    CheckMaterialScore;
  end;
end;

procedure GenerateLegalMoves(MyColour, YourColour: BoardColourEnum;
	var nMoves: ExtMoveRange; var Moves: MoveList);
var
	iMove, jMove: ExtMoveRange;
  CheckFlag: CheckEnum;
begin
	GenerateMoves(MyColour, YourColour, AllMoves, nMoves, Moves);
  iMove := 1;
  while iMove <= nMoves do
  begin
		DoMove(MyColour, YourColour, Moves[iMove]);
    CheckFlag := ReturnCheck(MyColour, YourColour);
		UndoMove(MyColour, YourColour, Moves[iMove]);
    if CheckFlag = Check then
    begin
    	for jMove := iMove to nMoves - 1 do
      	Moves[jMove] := Moves[jMove + 1];
      Dec(nMoves);
    end
    else
    	Inc(iMove);
  end;
end;

function MoveOk(nMoves: ExtMoveRange; var Moves: MoveList;
	TheMove: MoveRecord): Boolean;
var
	iMove: ExtMoveRange;
begin
	Result := True;
	for iMove := 1 to nMoves do
  begin
  	if Moves[iMove].AllFields = TheMove.AllFields then
    	Exit;
  end;
  Result := False;
end;

end.
