{
ZZZZZZ 5.0, a chess program.
Copyright (C) 1998 G. Wiesenekker
E-mail: gijsbert.wiesenekker@wxs.nl
http://home.wxs.nl/~gijsbert.wiesenekker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}
unit Search;

{$MODE Delphi}

interface

uses Chess;

function DrawByRepetition(Strict: Boolean): Boolean;
procedure DoSearch(MyColour, YourColour: BoardColourEnum;
	nMoves: ExtMoveRange; var Moves: MoveList);

implementation

uses Moves, Score, Cache, Timer, Main, SysUtils, LCLIntf, LCLType, LMessages, Thread, DebugBoard;

const
  QuiesDepth = 2;
var
	TimeCount: Integer;
	BestLine: AnsiString;
  RootScore: array [BoardColourEnum] of ScoreRecord;
  RootMaterialBalance: array [BoardColourEnum] of Integer;
  SearchScore, ProblemScore: ScoreRecord;

function DrawByRepetition(Strict: Boolean): Boolean;
var
	iNode: Integer;
  nRep: Integer;
begin
  iNode := Node - 2;
  nRep := 1;
  while iNode > 0 do
  begin
  	if NodeList[iNode].NodeCapture then
    	break;
  	if NodeList[iNode].NodeFullHashKey = NodeList[Node].NodeFullHashKey then
    begin
    	Inc(nRep);
      if (not Strict) or (nRep = 3) then
      begin
    		Result := True;
      	Exit;
      end;
    end;
    Dec(iNode, 2);
  end;
  Result := False;
end;

procedure QuiesLine(MyColour, YourColour: BoardColourEnum; MyMove: MoveRecord);
var
	iCache: Integer;
  nYourMoves: ExtMoveRange;
  YourMoves: MoveList;
begin
	if DrawByRepetition(False) then
  begin
  	BestLine := BestLine + ' Rep!';
		Exit;
  end;
  BestLine := BestLine + ' ' + MoveToLongString(MyMove);
	DoMove(MyColour, YourColour, MyMove);
  iCache := ReturnQuiesCacheIndex;
  with QuiesCache[YourColour][iCache] do
  begin
  	if (QuiesBestMove.MoveKind <> [Invalid]) and
       (QuiesCacheKey = NodeList[Node].NodeFullHashKey) then
    begin
{$IFDEF TraceScore}
    	BestLine := BestLine + ' q ' + ScoreToString(QuiesCacheScore);
{$ENDIF}
    	GenerateMoves(YourColour, MyColour, AllMoves, nYourMoves, YourMoves);
      if not MoveOk(nYourMoves, YourMoves, QuiesBestMove) then
      	raise EChessError.Create('Quies move not in list of moves');
      QuiesLine(YourColour, MyColour,  QuiesBestMove)
    end;
  end;
	UndoMove(MyColour, YourColour, MyMove);
end;

procedure AlphaBetaLine(MyColour, YourColour: BoardColourEnum; MyMove: MoveRecord);
var
	iCache: Integer;
  nYourMoves: ExtMoveRange;
  YourMoves: MoveList;
  Found: Boolean;
begin
	if DrawByRepetition(False) then
  begin
  	BestLine := BestLine + ' Rep!';
		Exit;
  end;
  BestLine := BestLine + ' ' + MoveToLongString(MyMove);
	DoMove(MyColour, YourColour, MyMove);
  iCache := ReturnCacheIndex;
  Found := False;
  with AlphaBetaCache[YourColour][iCache] do
  begin
  	if (CacheBestMove.MoveKind <> [Invalid]) and
       (CacheKey = NodeList[Node].NodeFullHashKey) then
    begin
    	Found := True;
{$IFDEF TraceScore}
      BestLine := BestLine + ' ab ' + ScoreToString(CacheScore);
{$ENDIF}
    	GenerateMoves(YourColour, MyColour, AllMoves, nYourMoves, YourMoves);
      //if not MoveOk(nYourMoves, YourMoves, CacheBestMove) then
      	//raise EChessError.Create('Alpha beta move not in list of moves');
      AlphaBetaLine(YourColour, MyColour, CacheBestMove)
    end;
  end;
  if not Found then
  begin
    iCache := ReturnQuiesCacheIndex;
    with QuiesCache[YourColour][iCache] do
    begin
      if (QuiesBestMove.MoveKind <> [Invalid]) and
         (QuiesCacheKey = NodeList[Node].NodeFullHashKey) then
      begin
{$IFDEF TraceScore}
      	BestLine := BestLine + ' q ' + ScoreToString(QuiesCacheScore);
{$ENDIF}
	    	GenerateMoves(YourColour, MyColour, AllMoves, nYourMoves, YourMoves);
      	if not MoveOk(nYourMoves, YourMoves, QuiesBestMove) then
      		raise EChessError.Create('Quies move not in list of moves');
        QuiesLine(YourColour, MyColour,  QuiesBestMove)
      end;
    end;
  end;
	UndoMove(MyColour, YourColour, MyMove);
end;

function QuiesSearch(MyColour, YourColour: BoardColourEnum; MaterialOnly: Boolean;
	MyAlpha, MyBeta: ScoreRecord; Depth: ExtNodeRange; CheckFlag: CheckEnum): ScoreRecord;
var
	nMoves: ExtMoveRange;
  NoMoves: Boolean;
  Moves: MoveList;
	BestScore: ScoreRecord;
  BestMove: MoveRecord;
  YourAlpha, YourBeta: ScoreRecord;
  TrueScore, MyScore, YourScore: ScoreRecord;
	iMove: Integer;
  iCache: Integer;
label
	LabelReturn, LabelStore;
begin
	Inc(Nodes);
	if CompareScore(MyAlpha, GE, MyBeta) then
  	raise EChessError.Create('Alpha >= Beta');
  if DrawByRepetition(False) then
  begin
  	BestScore := DrawScore;
    goto LabelReturn;
  end;

	BestMove := InvalidMove;
  iCache := ReturnQuiesCacheIndex;

	if CheckFlag = CheckUnknown then
  	CheckFlag := ReturnCheck(MyColour, YourColour);
  Dangerous := False;
  if CheckFlag = NoCheck then
  begin
  	Inc(Depth);
    if Options.UseLazy and (MaterialScore[MyColour] >= 10 * MaterialValue[Pawn]) then
    begin
      MyScore.MatScore :=
        MaterialScore[MyColour] - MaterialScore[YourColour];
      MyScore.PosScore := -InfiniteScore;
      MyScore.VisitScore := -InfiniteScore;
      if CompareScore(MyScore, GE, MyBeta) then
      begin
        BestScore := MyScore;
        goto LabelReturn;
      end;
  	end;
    if MaterialOnly then
    begin
    	TrueScore.MatScore := MaterialScore[MyColour] - MaterialScore[YourColour];
      TrueScore.PosScore := 0;
      TrueScore.VisitScore := 0;
      Dangerous := False;
    end
    else
			SetScore(MyColour, YourColour, TrueScore);
    if not Dangerous then
    begin
      BestScore := MyAlpha;
      if CompareScore(TrueScore, GT, BestScore) then
      begin
        BestScore := TrueScore;
        if CompareScore(BestScore, GE, MyBeta) then
          goto LabelReturn;
      end;
      GenerateMoves(MyColour, YourColour, CapturesOnly, nMoves, Moves);
    end
    else
    begin
  		GenerateMoves(MyColour, YourColour, AllMoves, nMoves, Moves);
	  	BestScore := MyAlpha;
    end;
  end
  else
  begin
  	GenerateMoves(MyColour, YourColour, AllMoves, nMoves, Moves);
  	BestScore := MyAlpha;
  end;
  NoMoves := True;
  for iMove := 1 to nMoves do
  begin
  	if CheckFlag = NoCheck then
    begin
  		if Depth >= QuiesDepth then
    	begin
				if not Board[Moves[iMove].MoveTo].ListCaptured then
      		Continue;
    	end;
    end;
  	DoMove(MyColour, YourColour, Moves[iMove]);
    if ReturnCheck(MyColour, YourColour) = Check then
    begin
    	UndoMove(MyColour, YourColour, Moves[iMove]);
      Continue;
    end;
    NoMoves := False;
		FlipScore(BestScore, YourBeta);
    FlipScore(MyBeta, YourAlpha);
  	YourScore := QuiesSearch(YourColour, MyColour, MaterialOnly,
    	YourAlpha, YourBeta, Depth + 1, CheckUnknown);
    UndoMove(MyColour, YourColour, Moves[iMove]);
    if CompareScore(YourScore, EQ, MinusInfinity) then
    begin
    	BestScore := YourScore;
    	goto LabelReturn;
    end;
    FlipScore(YourScore, MyScore);
    if CompareScore(MyScore, GT, BestScore) then
    begin
    	BestScore := MyScore;
      BestMove := Moves[iMove];
      if CompareScore(BestScore, GE, MyBeta) then
      	goto LabelStore;
    end;
  end;
  if NoMoves then
  begin
  	if CheckFlag = Check then
  	begin
    	BestScore := LostScore;
    	Inc(BestScore.MatScore, Node);
    end
    else if Dangerous then
    begin
    	BestScore := DrawScore;
    end
  end;
  LabelStore:
  if not MaterialOnly and (BestMove.MoveKind <> [Invalid]) then
  begin
  	with QuiesCache[MyColour][iCache] do
  	begin
  		QuiesCacheKey := NodeList[Node].NodeFullHashKey;
    	QuiesBestMove := BestMove;
{$IFDEF TraceScore}
      QuiesCacheScore := BestScore;
{$ENDIF}
  	end;
  end;
  LabelReturn:
  Result := BestScore;
end;

function AlphaBeta(MyColour, YourColour: BoardColourEnum;
	MyAlpha: ScoreRecord; MyBeta: ScoreRecord; MyDepth: ExtNodeRange; AllowNull: Boolean): ScoreRecord;
var
	BestScore: ScoreRecord;
  BestMove: MoveRecord;
  nMoves: ExtMoveRange;
  NoMoves: Boolean;
  Moves: MoveList;
	YourDepth, NullDepth: Integer;
	YourAlpha, YourBeta: ScoreRecord;
  CheckAlpha, CheckBeta: ScoreRecord;
  MyScore, YourScore: ScoreRecord;
  iMove, jMove: Integer;
  CheckFlag: CheckEnum;
  iCache: Integer;
  TempStatus: SearchStatusEnum;
label
	LabelStore, LabelReturn;
begin
	if CompareScore(MyAlpha, GE, MyBeta) then
  	raise EChessError.Create('Alpha >= Beta');
  if DrawByRepetition(False) then
  begin
  	BestScore := DrawScore;
    goto LabelReturn;
  end;

  BestScore := MyAlpha;
  BestMove := InvalidMove;

  Inc(TimeCount);
  if TimeCount = 100 then
  begin
    TimeCount := 0;
    if UpdateTimeUsed then
    begin
      if ReturnTimeUsed >= TimeSearch then
      begin
      	if CompareScore(SearchScore, LT, ProblemScore) then
        begin
        	if TimeProblem > 0 then
          begin
          	SearchThread.MyPrintf('losing material, using more time..');
            Inc(TimeSearch, TimeProblem);
            TimeProblem := 0;
          end;
        end;
      end;
      if ReturnTimeUsed >= TimeSearch then
        SearchThread.SetSearchStatus(SearchStopping);
      TempStatus := SearchThread.ReadSearchStatus;
      If (TempStatus = SearchStopping) or (TempStatus = SearchAborting) then
      begin
        BestScore := MinusInfinity;
        goto LabelReturn;
      end;
	    SearchThread.UpdateNodes;
    end;
  end;

  iCache := ReturnCacheIndex;
  with AlphaBetaCache[MyColour][iCache] do
  begin
  	if Options.UseCache and (NodeList[Node].NodeFullHashKey = CacheKey) then
    begin
    	BestMove := CacheBestMove;
      if CacheDepth >= MyDepth then
      begin
      	if CacheKind = TrueScore then
        begin
        	BestScore := CacheScore;
          goto LabelReturn;
        end;
        if (CacheKind = LowerBound) and CompareScore(CacheScore, GT, MyAlpha) then
          MyAlpha := CacheScore
        else if (CacheKind = UpperBound) and CompareScore(CacheScore, LT, MyBeta) then
          MyBeta := CacheScore;
        BestScore := MyAlpha;
        if CompareScore(BestScore, GE, MyBeta) then
          goto LabelReturn;
      end;
    end;
  end;

  CheckFlag := ReturnCheck(MyColour, YourColour);
  if CheckFlag = NoCheck then
  begin
		if MyDepth <= 0 then
  	begin
  		BestScore := QuiesSearch(MyColour, YourColour, False, MyAlpha, MyBeta, 0, CheckFlag);
    	goto LabelReturn;
  	end;
  	YourDepth := MyDepth - 1;
    if Options.UseNullMove and AllowNull and
    	 (MaterialScore[MyColour] >= 10 * MaterialValue[Pawn]) and
       CompareScore(MyBeta, LT, PlusInfinity) then
    begin
      if YourDepth <= 1 then
      begin
  			MyScore := QuiesSearch(MyColour, YourColour, False, MyAlpha, MyBeta, 0, CheckFlag);
				if CompareScore(MyScore, GE, MyBeta) then
        begin
          BestScore := MyScore;
          goto LabelStore;
        end;
      end
      else
    	begin
      	if NodeList[Node - 1].NodeNull then
        	raise EChessError.Create('Duplicate null move');
        FlipScore(MyBeta, YourAlpha);
        FlipScore(MyBeta, YourBeta);
        Inc(YourBeta.VisitScore);
        DoMove(MyColour, YourColour, NullMove);
        NullDepth := YourDepth - 1;
        if Odd(NullDepth) then
      		Dec(NullDepth);
        YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta, NullDepth, False);
        UndoMove(MyColour, YourColour, NullMove);
        if CompareScore(YourScore, EQ, MinusInfinity) then
        begin
          BestScore := YourScore;
          goto LabelReturn;
        end;
        FlipScore(YourScore, MyScore);
        if CompareScore(MyScore, GE, MyBeta) then
        begin
          BestScore := MyScore;
          BestMove := NullMove;
          goto LabelStore;
        end;
      end;
    end;
  end
  else
  begin
    if Options.ExtendCheck or (not NodeList[Node - 1].NodeCapture)then
    begin
    	YourDepth := MyDepth;
    end
    else
    begin
    	//do not extend checks that cost material
      CheckBeta := RootScore[MyColour];
      Inc(CheckBeta.MatScore, MaterialValue[Pawn] div 2);
      CheckAlpha := CheckBeta;
      Dec(CheckAlpha.MatScore);
      MyScore := QuiesSearch(MyColour, YourColour, True, CheckAlpha, CheckBeta, 0, CheckFlag);
      if CompareScore(MyScore, LE, CheckAlpha) then
      begin
      	YourDepth := MyDepth;
      end
      else
      begin
        if MyDepth <= 0 then
        begin
          BestScore := QuiesSearch(MyColour, YourColour, False, MyAlpha, MyBeta, 0, CheckFlag);
          goto LabelReturn;
        end;
        YourDepth := MyDepth - 1;
      end;
    end;
  end;

	GenerateMoves(MyColour, YourColour, AllMoves, nMoves, Moves);
  if (BestMove.MoveKind <> [Invalid]) and (BestMove.MoveKind <> [Null]) then
  begin
  	jMove := Low(ExtMoveRange);
    for iMove := 1 to nMoves do
    begin
    	if BestMove.AllFields = Moves[iMove].AllFields then
      begin
      	jMove := iMove;
        break;
      end;
    end;
    if jMove <> Low(ExtMoveRange) then
    begin
    	for iMove := jMove downto 2 do
      	Moves[iMove] := Moves[iMove - 1];
      Moves[1] := BestMove;
    end
    else
    	raise EChessError.Create('Best move not in list of legal moves');
  end;
  NoMoves := True;
	for iMove := 1 to nMoves do
  begin
  	DoMove(MyColour, YourColour, Moves[iMove]);
    if ReturnCheck(MyColour, YourColour) = Check then
    begin
    	UndoMove(MyColour, YourColour, Moves[iMove]);
      Continue;
    end;
    NoMoves := False;
		FlipScore(BestScore, YourBeta);
    FlipScore(MyBeta, YourAlpha);
    YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta, YourDepth, True);
    UndoMove(MyColour, YourColour, Moves[iMove]);
    if CompareScore(YourScore, EQ, MinusInfinity) then
    begin
    	BestScore := YourScore;
    	goto LabelReturn;
    end;
    FlipScore(YourScore, MyScore);
    if CompareScore(MyScore, GT, BestScore) then
    begin
    	BestScore := MyScore;
      BestMove := Moves[iMove];
      if CompareScore(BestScore, GE, MyBeta) then
      	goto LabelStore;
    end;
  end;
  if NoMoves then
  begin
  	if CheckFlag = Check then
    begin
  		BestScore := LostScore;
      //better if mate further away
      Inc(BestScore.MatScore, Node);
    end
    else
    	BestScore := DrawScore;
  end;
  LabelStore:
  if BestMove.MoveKind <> [Invalid] then
  begin
  	if not CompareScore(BestScore, EQ, DrawScore) then
    begin
      with AlphaBetaCache[MyColour][iCache] do
      begin
        CacheKey := NodeList[Node].NodeFullHashKey;
        CacheBestMove := BestMove;
        CacheScore := BestScore;
        CacheDepth := MyDepth;
        if CompareScore(BestScore, LE, MyAlpha) then
          CacheKind := UpperBound
        else if CompareScore(BestScore, GE, MyBeta) then
          CacheKind := LowerBound
        else
          CacheKind := TrueScore;
      end;
    end;
  end;
  LabelReturn:
	Result := BestScore;
end;

procedure DoSearch(MyColour, YourColour: BoardColourEnum;
	nMoves: ExtMoveRange; var Moves: MoveList);
var
	MyAlpha, MyBeta: ScoreRecord;
  YourAlpha, YourBeta: ScoreRecord;
  MyDepth, YourDepth: ExtNodeRange;
  MyScore, YourScore: ScoreRecord;
  iMove, jMove: Integer;
	TempMove: MoveRecord;
label
	LabelBreak;
begin
	ResetTimeUsed;
  ResetCacheDepth;
  Nodes := 0;
  TimeCount := 0;
  ShowScore := False;
  SetScore(MyColour, YourColour, SearchScore);
  ProblemScore := SearchScore;
  Dec(ProblemScore.MatScore, MaterialValue[Pawn] * 3 div 4);
	RootScore[MyColour].MatScore := MaterialScore[MyColour] - MaterialScore[YourColour];
  RootScore[MyColour].PosScore := 0;
  RootScore[MyColour].VisitScore := 0;
  RootMaterialBalance[MyColour] := MaterialScore[MyColour] - MaterialScore[YourColour];
  RootMaterialBalance[YourColour] := -RootMaterialBalance[MyColour];
  FlipScore(RootScore[MyColour], RootScore[YourColour]);
  for MyDepth := 1 to Options.DepthLimit do
  begin
  	YourDepth := MyDepth - 1;
  	MyAlpha := SearchScore;
    Dec(MyAlpha.MatScore, MaterialValue[Pawn] div 2);
    MyBeta := SearchScore;
    Inc(MyBeta.MatScore, MaterialValue[Pawn] div 2);
    FlipScore(MyAlpha, YourBeta);
    FlipScore(MyBeta, YourAlpha);
  	DoMove(MyColour, YourColour, Moves[1]);
  	YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta,
    	YourDepth, True);
    UndoMove(MyColour, YourColour, Moves[1]);
    if CompareScore(YourScore, EQ, MinusInfinity) then
    	goto LabelBreak;
    FlipScore(YourScore, MyScore);
    if CompareScore(MyScore, LE, MyAlpha) then
    begin
	  	DoMove(MyColour, YourColour, Moves[1]);
    	MyAlpha := MinusInfinity;
      MyBeta := MyScore;
    	FlipScore(MyAlpha, YourBeta);
    	FlipScore(MyBeta, YourAlpha);
  		YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta, YourDepth, True);
	    UndoMove(MyColour, YourColour, Moves[1]);
    	if CompareScore(YourScore, EQ, MinusInfinity) then
    		goto LabelBreak;
      FlipScore(YourScore, SearchScore);
    end
    else if CompareScore(MyScore, GE, MyBeta) then
    begin
	  	DoMove(MyColour, YourColour, Moves[1]);
    	MyAlpha := MyScore;
      MyBeta := PlusInfinity;
    	FlipScore(MyAlpha, YourBeta);
    	FlipScore(MyBeta, YourAlpha);
  		YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta, YourDepth, True);
	    UndoMove(MyColour, YourColour, Moves[1]);
    	if CompareScore(YourScore, EQ, MinusInfinity) then
    		goto LabelBreak;
      FlipScore(YourScore, SearchScore);
    end
    else
    	SearchScore := MyScore;
    BestLine := IntToStr(MyDepth);
    AlphaBetaLine(MyColour, YourColour, Moves[1]);
    BestLine := BestLine + ':'+ ScoreToString(SearchScore);
    SearchThread.MyPrintf(BestLine);
    for iMove := 2 to nMoves do
    begin
    	MyAlpha := SearchScore;
      MyBeta := SearchScore;
      Inc(MyBeta.VisitScore);
      FlipScore(MyAlpha, YourBeta);
      FlipScore(MyBeta, YourAlpha);
	  	DoMove(MyColour, YourColour, Moves[iMove]);
  		YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta, YourDepth, True);
	    UndoMove(MyColour, YourColour, Moves[iMove]);
    	if CompareScore(YourScore, EQ, MinusInfinity) then
    		goto LabelBreak;
      FlipScore(YourScore, MyScore);
      if CompareScore(MyScore, GE, MyBeta) then
      begin
      	TempMove := Moves[iMove];
        for jMove := iMove downto 2 do
        	Moves[jMove] := Moves[jMove - 1];
        Moves[1] := TempMove;
        BestLine := IntToStr(MyDepth) + ' ' + MoveToLongString(Moves[1]);
    		SearchThread.MyPrintf(BestLine);
        DoMove(MyColour, YourColour, Moves[1]);
    		MyAlpha := MyScore;
      	MyBeta := PlusInfinity;
    		FlipScore(MyAlpha, YourBeta);
    		FlipScore(MyBeta, YourAlpha);
        YourScore := AlphaBeta(YourColour, MyColour, YourAlpha, YourBeta, YourDepth, True);
        UndoMove(MyColour, YourColour, Moves[1]);
        if CompareScore(YourScore, EQ, MinusInfinity) then
          goto LabelBreak;
        FlipScore(YourScore, SearchScore);
        BestLine := IntToStr(MyDepth);
        AlphaBetaLine(MyColour, YourColour, Moves[1]);
        BestLine := BestLine + ':'+ ScoreToString(SearchScore);
        SearchThread.MyPrintf(BestLine);
      end;
    end;
    if CompareScore(SearchScore, GE, WonLowerBound) then
     Break;
    if CompareScore(SearchScore, LE, LostUpperBound) then
     Break;
  end;
	LabelBreak:
  SearchThread.SetSearchStatus(SearchStopping);
  SearchThread.MyPrintf('Search ready');
end;

end.
