{
ZZZZZZ 5.0, a chess program.
Copyright (C) 1998 G. Wiesenekker
E-mail: gijsbert.wiesenekker@wxs.nl
http://home.wxs.nl/~gijsbert.wiesenekker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}
unit Thread;

{$MODE Delphi}

interface

uses
  Classes, Chess;

type
	SearchStatusEnum = (SearchStarting, AppendStarting, SearchStarted, SearchStopping, SearchAborting, SearchStopped);
  TSearchThread = class(TThread)
  private
    { Private declarations }
    SynchronizeLine: AnsiString;
    SynchronizeSearchStatus: SearchStatusEnum;
  	procedure SynchronizeReadSearchStatus;
  	procedure SynchronizeSetSearchStatus;
    procedure SynchronizeMyPrintf;
    procedure SynchronizeThreadResult;
    procedure SynchronizeNodes;
    procedure SynchronizeShowBoard;
  protected
    procedure Execute; override;
  public
    procedure MyPrintf(Line: AnsiString);
    function ReadSearchStatus: SearchStatusEnum;
    procedure SetSearchStatus(TheStatus: SearchStatusEnum);
    procedure UpdateNodes;
    procedure ShowBoard;
  end;
var
	SearchStatus: SearchStatusEnum;
	SearchThread: TSearchThread;
  MyThreadColour, YourThreadColour: BoardColourEnum;
	nThreadMoves: ExtMoveRange;
  ThreadMoves: MoveList;
	Nodes: Integer;
procedure ChangeSearchStatus(TheStatus: SearchStatusEnum);

implementation

uses Main, LCLIntf, LCLType, LMessages, Moves, Search, DebugBoard, SysUtils;

var
	MoveTimePrevious, MoveTimeCurrent: TDateTime;

{ Important: Methods and properties of objects in VCL can only be used in a
  method called using Synchronize, for example,

      Synchronize(UpdateCaption);

  and UpdateCaption could look like,

    procedure SearchThread.UpdateCaption;
    begin
      Form1.Caption := 'Updated in a thread';
    end; }

{ SearchThread }

procedure ChangeSearchStatus(TheStatus: SearchStatusEnum);
begin
	if (SearchStatus = SearchStopped) and
  	 ((TheStatus = SearchStarting) or (TheStatus = AppendStarting)) then
  begin
  	SearchStatus := TheStatus;
  end
  else if ((SearchStatus = SearchStarting) or (SearchStatus = AppendStarting)) and
		      (TheStatus = SearchStarted) then
  begin
  	SearchStatus := TheStatus;
  end
  else if (SearchStatus = SearchStarted) and
  	      ((TheStatus = SearchStopping) or (TheStatus = SearchAborting)) then
  begin
  	SearchStatus := TheStatus;
  end
  else if ((SearchStatus = SearchStopping) or (SearchStatus = SearchAborting)) and
  	      (TheStatus = SearchStopped) then
  begin
  	SearchStatus := TheStatus;
  end;
end;

procedure TSearchThread.SynchronizeReadSearchStatus;
begin
	SynchronizeSearchStatus := SearchStatus;
end;

function TSearchThread.ReadSearchStatus: SearchStatusEnum;
begin
	Synchronize(SynchronizeReadSearchStatus);
  Result := SynchronizeSearchStatus;
end;

//

procedure TSearchThread.SynchronizeSetSearchStatus;
begin
  ChangeSearchStatus(SynchronizeSearchStatus);
end;

procedure TSearchThread.SetSearchStatus(TheStatus: SearchStatusEnum);
begin
	SynchronizeSearchStatus := TheStatus;
  Synchronize(SynchronizeSetSearchStatus);
end;

procedure TSearchThread.SynchronizeMyPrintf;
begin
	MainForm.OutputMemo.Lines.Add(SynchronizeLine);
  UpdateWindows;
end;

procedure TSearchThread.SynchronizeThreadResult;
var
	MoveTime: Integer;
begin
	MainForm.OutputMemo.Lines.Add(MoveToLongString(ThreadMoves[1]));
	SetSearchStatus(SearchStopped);
  MoveTime := Trunc((MoveTimeCurrent - MoveTimePrevious) * SecsPerDay);
  MainForm.OutputMemo.Lines.Add('used ' +  IntToStr(MoveTime));
  Inc(GameTimeUsed, MoveTime);
	HandleMoveInput(MoveToLongString(ThreadMoves[1]));
end;

procedure TSearchThread.SynchronizeNodes;
begin
	//psychological warfare
	MainForm.NodesEdit.Text := IntToStr(9 * Nodes);
end;

procedure TSearchThread.SynchronizeShowBoard;
begin
	DebugBoardForm.ShowModal;
end;

procedure TSearchThread.UpdateNodes;
begin
  Synchronize(SynchronizeNodes);
end;

procedure TSearchThread.ShowBoard;
begin
	Synchronize(SynchronizeShowBoard);
end;

procedure TSearchThread.MyPrintf(Line: AnsiString);
begin
	SynchronizeLine := Line;
  Synchronize(SynchronizeMyPrintf);
end;

procedure TSearchThread.Execute;
var
	TempStatus: SearchStatusEnum;
begin
  { Place thread code here }
  MyPrintf('Thread started');
  while True do
  begin
  	Sleep(100);
    if Terminated then
    	Break;
    TempStatus := ReadSearchStatus;
    if TempStatus = SearchStarting then
    begin
    	MyPrintf('Search starting');
      SetSearchStatus(SearchStarted);
    	MyPrintf('Search started');
			DoSearch(MyThreadColour, YourThreadColour, nThreadMoves, ThreadMoves);
      SetSearchStatus(SearchStopped);
    end
    else if TempStatus = AppendStarting then
    begin
    	MyPrintf('Append starting');
      SetSearchStatus(SearchStarted);
    	MyPrintf('Append started');
      MoveTimePrevious := Time;
      if nThreadMoves > 1 then
				DoSearch(MyThreadColour, YourThreadColour, nThreadMoves, ThreadMoves)
      else
      	SetSearchStatus(SearchStopping);
      MoveTimeCurrent := Time;
      if ReadSearchStatus = SearchStopping then
				SynchronizeThreadResult;
      SetSearchStatus(SearchStopped);
    end
    else if TempStatus = SearchStopping then
    begin
    	//only dummy
    	MyPrintf('Search stopping');
      SetSearchStatus(SearchStopped);
      MyPrintf('Search stopped');
    end
    else if TempStatus = SearchAborting then
    begin
    	//only dummy
    	MyPrintf('Search aborted');
      SetSearchStatus(SearchStopped);
      MyPrintf('Search stopped');
    end
  end;
end;

end.
